@@ -1,8 +1,8 @@
-#!/usr/bin/env python
 """
 This module provides access to the API of a foreman server
 """
 import re
+import six
 import json
 import copy
 import types
@@ -27,6 +27,17 @@ if requests.__version__.split('.', 1)[0] == '0':
 else:
     OLD_REQ = False
 
+logger = logging.getLogger(__name__)
+
+
+def set_loglevel(level):
+    """
+    Sets the loglevel for the python-foreman module.
+
+    :param loglevel: a loglevel constant from the logging module.
+    """
+    logger.setLevel(level)
+
 
 def try_int(what):
     try:
@@ -105,6 +116,7 @@ class ForemanVersionException(Exception):
 
 class MethodAPIDescription(object):
     exclude_html_reg = re.compile('</?[^>/]+/?>')
+    resource_pattern = re.compile(r'^/api(/v[12])?/(?P<resource>\w+).*')
 
     def __init__(self, resource, method, api):
         self._method = copy.deepcopy(method)
@@ -113,11 +125,7 @@ class MethodAPIDescription(object):
         self.url = self._api['api_url']
         self.url_params = re.findall('/:([^/]+)(?:/|$)', self.url)
         self.params = self._method['params']
-        # special case for the api root
-        if self.url == '/api':
-            self.resource = 'api'
-        else:
-            self.resource = self.url.split('/', 3)[2]
+        self.resource = self.parse_resource_from_url(self.url) or ''
         self.name = self._get_name()
         self.http_method = self._api['http_method']
         self.short_desc = self._api['short_description'] or ''
@@ -125,6 +133,21 @@ class MethodAPIDescription(object):
     def __repr__(self):
         return "<resource:%s, name:%s>" % (self.resource, self.name)
 
+    def parse_resource_from_url(self, url):
+        """
+        Returns the appropriate resource name for the given URL.
+
+        :param url:  API URL stub, like: '/api/hosts'
+        :return: Resource name, like 'hosts', or None if not found
+        """
+        # special case for the api root
+        if url == '/api':
+            return 'api'
+
+        match = self.resource_pattern.match(url)
+        if match:
+            return match.groupdict().get('resource', None)
+
     def _get_name(self):
         """
         There are three cases, because apipie definitions can have multiple
@@ -200,7 +223,7 @@ class MethodAPIDescription(object):
         # split required and non-required params for the definition
         req_params = []
         nonreq_params = []
-        for param in params.itervalues():
+        for param in six.itervalues(params):
             if param['required']:
                 req_params.append(param)
             else:
@@ -251,7 +274,7 @@ class MethodAPIDescription(object):
 
         code = '\n'.join(code)
 
-        exec code
+        six.exec_(code)
 
         function = locals()[self.name]
         # to ease debugging, all the funcs have the definitions attached
@@ -297,7 +320,7 @@ def parse_resource_definition(resource_name, resource_dct):
         '_conflicting_methods': [],
     }
 
-    # methods in foreign_methods are ment for other resources,
+    # methods in foreign_methods are meant for other resources,
     # that is, the url and the resource field do not match /api/{resource}
     foreign_methods = {}
 
@@ -313,7 +336,7 @@ def parse_resource_definition(resource_name, resource_dct):
             api = MethodAPIDescription(resource_name, method, api)
 
             if api.resource != resource_name:
-                # this means that the json apipie passes says that an
+                # this means that the json apipie passed says that an
                 # endpoint in the form: /api/{resource}/* belongs to
                 # {different_resource}, we just put it under {resource}
                 # later, storing it under _foreign_methods for now as we
@@ -321,8 +344,13 @@ def parse_resource_definition(resource_name, resource_dct):
                 functions = foreign_methods.setdefault(api.resource, {})
                 if api.name in functions:
                     old_api = functions.get(api.name).defs
-                    logging.warning(
-                        "There is conflict trying to redefine a method "
+                    # show only in debug the repeated but identical definitions
+                    log_method = logger.warning
+                    if api.url == old_api.url:
+                        log_method = logger.debug
+
+                    log_method(
+                        "There is a conflict trying to redefine a method "
                         "for a foreign resource (%s): \n"
                         "\tresource:\n"
                         "\tapipie_resource: %s\n"
@@ -345,8 +373,13 @@ def parse_resource_definition(resource_name, resource_dct):
                 # it's an own method, resource and url match
                 if api.name in new_dict['_own_methods']:
                     old_api = new_dict.get(api.name).defs
-                    logging.warning(
-                        "There is conflict trying to redefine method "
+                    log_method = logger.warning
+                    # show only in debug the repeated but identical definitions
+                    if api.url == old_api.url:
+                        log_method = logger.debug
+
+                    log_method(
+                        "There is a conflict trying to redefine method "
                         "(%s): \n"
                         "\tapipie_resource: %s\n"
                         "\tnew_api: %s\n"
@@ -434,17 +467,17 @@ class MetaForeman(type):
                     ['DEFS'],
                 )
             except ImportError:
-                logging.error('Unable to import plugin module %s', plugin)
+                logger.error('Unable to import plugin module %s', plugin)
                 continue
-            for http_method, funcs in myplugin.DEFS.iteritems():
+            for http_method, funcs in six.iteritems(myplugin.DEFS):
                 methods = MetaForeman.convert_plugin_def(http_method, funcs)
                 entries['methods'].extend(methods)
 
         def _init(self, foreman):
             super(self.__class__, self).__init__(foreman)
-            for name, value in self.__class__.__dict__.iteritems():
+            for name, value in six.iteritems(self.__class__.__dict__):
                 if isinstance(value, types.FunctionType) and name[0] != '_':
-                    logging.debug(
+                    logger.debug(
                         'Registering plugin method %s',
                         name,
                     )
@@ -453,8 +486,8 @@ class MetaForeman(type):
         resource_data, foreigns = parse_resource_definition('plugins', entries)
         # by default, all the methods are detected as foreign, we must manually
         # add them
-        for mname, method_dict in foreigns.iteritems():
-            resource_data[mname] = method_dict.values()[0]
+        for mname, method_dict in six.iteritems(foreigns):
+            resource_data[mname] = tuple(method_dict.values())[0]
             resource_data['_own_methods'].add(mname)
 
         plugins_cls = ResourceMeta.__new__(
@@ -478,12 +511,12 @@ class MetaForeman(type):
         """
         methods = []
         if http_method not in ('GET', 'PUT', 'POST', 'DELETE'):
-            logging.error(
+            logger.error(
                 'Plugin load failure, HTTP method %s unsupported.',
                 http_method,
             )
             return methods
-        for fname, params in funcs.iteritems():
+        for fname, params in six.iteritems(funcs):
             method = {
                 'apis': [{'short_description': 'no-doc'}],
                 'params': [],
@@ -491,7 +524,7 @@ class MetaForeman(type):
             method['apis'][0]['http_method'] = http_method
             method['apis'][0]['api_url'] = '/api/' + fname
             method['name'] = fname
-            for pname, pdef in params.iteritems():
+            for pname, pdef in six.iteritems(params):
                 param = {
                     'name': pname,
                     'validator': "Must be %s" % pdef['ptype'],
@@ -503,12 +536,12 @@ class MetaForeman(type):
         return methods
 
 
+@six.add_metaclass(MetaForeman)
 class Foreman(object):
     """
     Main client class. It's methods will be autogenerated, check the API docs
     for your foreman version `here <http://theforeman.org/api.html>`_.
     """
-    __metaclass__ = MetaForeman
 
     def __init__(self, url, auth=None, version=None, api_version=None,
                  use_cache=True, strict_cache=True, timeout=60,
@@ -523,7 +556,7 @@ class Foreman(object):
             will try to get them from the remote Foreman instance (it needs
             you to have disabled use_cache in the apipie configuration in your
             foreman instance)
-        :param sctrict_cache: If True, will not use a similar version
+        :param strict_cache: If True, will not use a similar version
             definitions file
         :param timeout: Timeout in seconds for each http request (default 60)
             If None or 0, then no timeout.
@@ -542,14 +575,14 @@ class Foreman(object):
         """
         if api_version is None:
             api_version = 1
-            logging.warning(
+            logger.warning(
                 "Api v1 will not be the default in the next version, if you "
                 "still want to use it, change the call to explicitly ask for "
                 "it. Though we recommend using the new and improved version 2"
             )
         self.url = url
         self._req_params = {}
-        self.timeout = {'DEFAULT': timeout}
+        self.timeout = {'DEFAULT': timeout or None}
 
         if timeout_post is not None:
             self.set_timeout(timeout_post, 'POST')
@@ -614,13 +647,26 @@ class Foreman(object):
         main page and extract the version from the footer.
         """
         params = dict(self._req_params)
-        home_page = self.session.get(self.url, **params)
-        match = re.search(r'Version\s+(?P<version>\S+)', home_page.text)
+        home_page = requests.get(
+            self.url,
+            verify=self.session.verify,
+            timeout=self.get_timeout('GET'),
+            **params
+        )
+
+        match = re.search(
+            r'Version\s+(?P<version>[^\s<]+)?',
+            home_page.text,
+        )
         if match:
             return match.groupdict()['version']
         else:
             # on newer versions the version can be taken from the status page
-            res = self.session.get(self.url + '/api/status', **params)
+            res = self.session.get(
+                self.url + '/api/status',
+                timeout=self.get_timeout('GET'),
+                **params
+            )
             if res.status_code < 200 or res.status_code >= 300:
                 raise ForemanException(
                     res,
@@ -641,22 +687,26 @@ class Foreman(object):
              minor version if no total match found
         """
         version = parse_version(self.version)
-        defs_path = os.path.join(os.path.dirname(__file__), 'definitions')
-        files = glob.glob('%s/*-v%s.json' % (defs_path, self.api_version))
-        files_version = [
-            (fn, parse_version(os.path.basename(fn).rsplit('-', 1)[0]))
-            for fn in files
-        ]
+        for cache_dir in [
+            os.path.join(os.path.expanduser('~'), '.python-foreman'),
+            os.path.dirname(__file__)
+        ]:
+            defs_path = os.path.join(cache_dir, 'definitions')
+            files = glob.glob('%s/*-v%s.json' % (defs_path, self.api_version))
+            files_version = [
+                (fn, parse_version(os.path.basename(fn).rsplit('-', 1)[0]))
+                for fn in files
+            ]
 
-        last_major_match = None
-        for f_name, f_ver in sorted(files_version, key=lambda x: x[1]):
-            if f_ver == version:
-                logging.debug('Found local cached version %s' % f_name)
-                return json.loads(open(f_name).read())
-            if f_ver[:2] == version[:2]:
-                last_major_match = f_name
-            if f_ver[0] > version[0]:
-                break
+            last_major_match = None
+            for f_name, f_ver in sorted(files_version, key=lambda x: x[1]):
+                if f_ver == version:
+                    logger.debug('Found local cached version %s' % f_name)
+                    return json.loads(open(f_name).read())
+                if f_ver[:2] == version[:2]:
+                    last_major_match = f_name
+                if f_ver[0] > version[0]:
+                    break
 
         if last_major_match:
             if strict:
@@ -666,7 +716,7 @@ class Foreman(object):
                     "without strict flag to use it"
                     % (self.version, last_major_match))
             else:
-                logging.warn(
+                logger.warn(
                     "Not exact version found, got cached %s for Foreman %s",
                     last_major_match,
                     self.version,
@@ -685,16 +735,27 @@ class Foreman(object):
 
     def _get_remote_defs(self):
         """
-        Retrieves the json definitions from remote forem
+        Retrieves the json definitions from remote foreman instance.
         """
         res = self.session.get(
             '%s/%s' % (self.url, 'apidoc/v%s.json' % self.api_version),
+            timeout=self.get_timeout('GET'),
             **self._req_params
         )
 
         if res.ok:
             data = json.loads(res.text)
-            defs_path = os.path.join(os.path.dirname(__file__), 'definitions')
+            defs_path = os.path.join(
+                os.path.expanduser('~'),
+                '.python-foreman',
+                'definitions'
+            )
+            if not os.path.exists(defs_path):
+                try:
+                    os.makedirs(defs_path)
+                except:
+                    logger.debug('Unable to create cache dir %s', defs_path)
+                    return data
             cache_fn = '%s/%s-v%s.json' % (
                 defs_path, self.version,
                 self.api_version,
@@ -702,11 +763,12 @@ class Foreman(object):
             try:
                 with open(cache_fn, 'w') as cache_fd:
                     cache_fd.write(json.dumps(data, indent=4, default=str))
+                    logger.debug('Wrote cache file %s', cache_fn)
             except:
-                logging.debug('Unable to write cache file %s', cache_fn)
+                logger.debug('Unable to write cache file %s', cache_fn)
         else:
             if res.status_code == 404:
-                logging.warn(
+                logger.warn(
                     "Unable to get api definition from live Foreman instance "
                     "at '%s', you might want to set the strict_cache to False."
                     "\nNOTE: Make sure that you have set the config.use_cache "
@@ -724,12 +786,12 @@ class Foreman(object):
         data = None
         if use_cache:
             try:
-                logging.debug("Trying local cached definitions first")
+                logger.debug("Trying local cached definitions first")
                 data = self._get_local_defs(strict=strict_cache)
             except ForemanVersionException as exc:
-                logging.debug(exc)
+                logger.debug(exc)
         if not data:
-            logging.debug("Checking remote server for definitions")
+            logger.debug("Checking remote server for definitions")
             data = self._get_remote_defs()
         return data
 
@@ -747,7 +809,7 @@ class Foreman(object):
 
         resource_defs = {}
         # parse all the defs first, as they may define methods cross-resource
-        for res_name, res_dct in data["docs"]["resources"].iteritems():
+        for res_name, res_dct in six.iteritems(data["docs"]["resources"]):
             new_resource, extra_foreign_methods = parse_resource_definition(
                 res_name.lower(),
                 res_dct,
@@ -757,7 +819,7 @@ class Foreman(object):
             # overwrites any methods
             if res_name in resource_defs:
                 old_res = resource_defs[res_name]
-                for prop_name, prop_val in new_resource.iteritems():
+                for prop_name, prop_val in six.iteritems(new_resource):
                     if (
                         prop_name == '_own_methods' and
                         prop_name in new_resource
@@ -768,7 +830,7 @@ class Foreman(object):
                     if prop_name.startswith('_'):
                         continue
                     if prop_name in old_res:
-                        logging.warning(
+                        logger.warning(
                             "There is conflict trying to redefine method "
                             "(%s) with foreign method: \n"
                             "\tapipie_resource: %s\n",
@@ -782,15 +844,15 @@ class Foreman(object):
 
             # update the other resources with the foreign methods, create
             # the resources if not there yet, merge if it already exists
-            for f_res_name, f_methods in extra_foreign_methods.iteritems():
+            for f_res_name, f_methods in six.iteritems(extra_foreign_methods):
                 methods = resource_defs.setdefault(
                     f_res_name,
                     {'_own_methods': set()},
                 )
 
-                for f_mname, f_method in f_methods.iteritems():
+                for f_mname, f_method in six.iteritems(f_methods):
                     if f_mname in methods:
-                        logging.warning(
+                        logger.warning(
                             "There is conflict trying to redefine method "
                             "(%s) with foreign method: \n"
                             "\tapipie_resource: %s\n",
@@ -803,7 +865,7 @@ class Foreman(object):
 
         # Finally ceate the resource classes for all the collected resources
         # instantiate and bind them to this class
-        for resource_name, resource_data in resource_defs.iteritems():
+        for resource_name, resource_data in six.iteritems(resource_defs):
             new_resource = ResourceMeta.__new__(
                 ResourceMeta,
                 str(resource_name),
@@ -811,7 +873,7 @@ class Foreman(object):
                 resource_data,
             )
             if not resource_data['_own_methods']:
-                logging.debug('Skiping empty resource %s' % resource_name)
+                logger.debug('Skipping empty resource %s' % resource_name)
                 continue
             instance = new_resource(self)
             setattr(self, resource_name, instance)
@@ -829,7 +891,7 @@ class Foreman(object):
             )
         try:
             return OLD_REQ and res.json or res.json()
-        except requests.JSONDecodeError:
+        except ValueError:
             return res.text
 
     def do_get(self, url, kwargs):
